#!/usr/bin/python

import sys
import random
import datetime
import traceback
from datetime import timedelta

from ardi.driver import histdriverbase
from ardi.driver import sqldriver
import pymssql

class ardihistdriver:
    def __init__(self):        
        self.port = 0
        self.db = None        
        self.host = ""
        self.database = ""
        self.username = ""
        self.password = ""
        self.query = ""
        self.stamp = ""
        self.table = ""
        self.lookup = ""
        self.value = ""
        self.filter = ""
        self.datestyle="1"
        self.tz = "local"
        self.options = ""
        self.lookuptable = {}

    def SetAddress(self, addr):
        self.connected = False
        bits = addr.split(":")
        filter = ""

        #print "Raw Address"

        if len(bits) >= 6:
            self.host = bits[0]
            self.port = bits[1]
            self.database = bits[2]
            self.username = bits[3]
            self.password = bits[4]
            self.table = bits[5]
            self.lookup = bits[6]
            self.stamp = bits[7]
            self.value = bits[8]
        if len(bits) > 9:
            self.filter = bits[9]
        if len(bits) > 11:
            self.datestyle = bits[10]
            self.tz = bits[11]
        if len(bits) > 12:
            self.options = bits[12]

    def Connect(self):
        if self.db is not None:
            try:
                self.Disconnect()
                self.db = None
            except:
                traceback.print_exc()
                pass
        
        try:
            self.core.logger.info("Connecting To " + self.host + ":" + self.port + " / " + self.database)
            self.db = pymssql.connect(self.host + ":" + self.port,self.username,self.password,self.database)
            self.connected = True
            self.core.logger.info("Successfully Connected!")
            self.db.close()
            return True
        except:
            return False        

    def Disconnect(self):
        try:            
            self.db.close()
        except:
            pass        
        
    def RunQuery(self, query):
        self.core.logger.info("Performing '" + query.function + "' query between " + str(query.sd) + " and " + str(query.ed) + " UTC")

        sd = query.sd
        ed = query.ed

        #Record query start time
        stime = datetime.datetime.now()

        #Connect to the database engine
        self.core.logger.info("Connecting To " + self.host + ":" + self.port + " / " + self.database)
        db = pymssql.connect(self.host + ":" + self.port,self.username,self.password,self.database)
        query.SetDateFormat(self.tz,self.datestyle)

        #Determine the function name for aggregation
        function = "AVG"
        startdate = str(query.sd)
        enddate = str(query.ed)

        #Place the date in quotes if it's a string-style date
        if self.datestyle == "1":
            startdate = "'" + startdate + "'"
            enddate = "'" + enddate + "'"

        #Prepare the 'Edge Query', used to pick up the previous and next values
        edgequery = "SELECT tagname,( select TOP 1 " + self.value + " from " + self.table + " WHERE " + self.table + "." + self.lookup + " = tags.tagname AND " + self.table + "." + self.stamp + " {cmp} ORDER BY " + self.stamp + " DESC) as value FROM "
        edgequery += "(VALUES ("
        any = False
        for add in query.addresses:
            if any == True:
                edgequery += ","
            edgequery += "'" + add + "'"
        edgequery += ")) as tags(tagname)"

        #Change the edge queries if this is a SNAPSHOT data source...
        snapshot = False
        if self.options == "snapshot":
            snapshot = True
            edgequery = 'SELECT ' + self.stamp + ', ' + self.lookup + ', ' + self.value + ' FROM ' + self.table + ' WHERE ' + self.stamp + ' = {dt}'
            if len(query.addresses) < 120:
                edgequery += " AND " + self.lookup + " IN ("
                any = False
                for add in query.addresses:
                    if any == True:
                        edgequery += ","
                    edgequery += "'" + add + "'"
                    any = True
                edgequery += ")"
        
        #Prepare the calculation that splits the times into periods in the main query
        if self.datestyle == "1":
            datecalc = "DATEADD(s,(CEILING(DATEDIFF(second," + startdate + "," + str(self.stamp) + ") / " + str(query.window) + ") * " + str(query.window) + ")," + startdate + ")"
        else:
            datecalc = "(CEILING((" + self.stamp + " - " + startdate + ") / " + str(query.window) + ") * " + str(query.window) + ") + " + startdate

        #Prepare the main query
        mainquery = "SELECT " + datecalc + "," + " " + self.lookup + "," + function + "(" + self.value + ") FROM " + self.table + " "
        mainquery += " WHERE ( " + self.stamp + " > " + startdate + " AND " + self.stamp + " < " + enddate + " ) "

        #If there are a reasonable number of tag names, add them into the query to filter it.
        if len(query.addresses) < 120:
            mainquery += "AND " + self.lookup + " IN ("
            any = False
            for add in query.addresses:
                if any == True:
                    mainquery += ","
                mainquery += "'" + add + "'"
                any = True
            mainquery += ")"
        mainquery += "GROUP BY " + datecalc + "," + self.lookup + " ORDER BY 1,2"

        timedpoints = 4

        #If the query is called as 'continue', it means we don't need to fetch the previous values.
        if query.function != 'continue':
            if snapshot == False:
                #Apply the upper edge query
                topoffquery = edgequery.replace("{cmp}","<=" + startdate)
                self.core.logger.debug("Upper Edge Query: " + topoffquery)
                try:
                    csr = db.cursor()
                    csr.execute(topoffquery);
                    for row in csr:
                        query.AddLine(str(row[0]),query.sd,row[1])
                except:
                    traceback.print_exc()
            else:
                #Get the most recent stamp BEFORE the start time...
                targetdate = ""
                topoffquery = 'SELECT TOP 1 ' + self.stamp + ' FROM ' + self.table + ' WHERE ' + self.stamp + " <= " + startdate + ' ORDER BY ' + self.stamp + ' DESC'
                self.core.logger.debug("Upper Edge Query #1: " + topoffquery)
                try:
                    csr = db.cursor()
                    csr.execute(topoffquery);
                    for row in csr:
                        targetdate = row[0]
                except:
                    traceback.print_exc()

                #print "Date: " + str(targetdate)
                #Polish off the edge query using the information picked up from above.
                topoffquery = edgequery
                vd = str(targetdate)
                if vd == "None":
                    NoDate = True
                else:
                    if self.datestyle == "1":
                        ps = vd.rfind('.')
                        if ps != -1:
                            if len(vd) > ps+timedpoints:
                                vd = vd[0:ps+timedpoints]

                        topoffquery = topoffquery.replace("{dt}","'" + vd + "'")
                    else:
                        topoffquery = topoffquery.replace("{dt}",vd)
                        
                #Now get all of the records that match
                self.core.logger.debug("Upper Edge Query #2: " + topoffquery)
                try:
                    csr = db.cursor()
                    csr.execute(topoffquery);
                    for row in csr:
                        query.AddLine(str(row[1]),row[0],row[2])
                except:
                    traceback.print_exc()

            elapsed = (datetime.datetime.now() - stime).total_seconds()

            self.core.logger.debug("Total Time After Upper Edge - " + str(elapsed))

        #Run the main query
        if sd != ed:
            try:
                csr = db.cursor()
                self.core.logger.debug("Main Query: " + mainquery)
                csr.execute(mainquery);
                for row in csr:
                    #print(str(row))
                    query.AddLine(str(row[1]),row[0],row[2])
            except:
                traceback.print_exc()

        elapsed = (datetime.datetime.now() - stime).total_seconds()

        self.core.logger.debug("Total Time After Content - " + str(elapsed))

        #Run the lower edge query
        if snapshot == False:
            topoffquery = edgequery.replace("{cmp}",">=" + enddate)
            self.core.logger.debug("Upper Edge Query: " + topoffquery)
            try:
                csr = db.cursor()
                csr.execute(topoffquery);
                for row in csr:
                    query.AddLine(str(row[0]),query.ed,row[1])
            except:
                traceback.print_exc()
        else:
            #Get the most recent stamp AFTER the start time...
            targetdate = ""
            topoffquery = 'SELECT TOP 1 ' + self.stamp + ' FROM ' + self.table + ' WHERE ' + self.stamp + " >= " + enddate + ' ORDER BY ' + self.stamp + ' ASC'
            self.core.logger.debug("Upper Edge Query #1: " + topoffquery)
            try:
                csr = db.cursor()
                csr.execute(topoffquery);
                for row in csr:
                    targetdate = row[0]
            except:
                traceback.print_exc()

            #Polish off the edge query using the information picked up from above.
            topoffquery = edgequery
            vd = str(targetdate)
            if vd == "None":
                NoDate = True
            else:
                if self.datestyle == "1":
                    ps = vd.rfind('.')
                    if ps != -1:
                        if len(vd) > ps + timedpoints:
                            vd = vd[0:ps + timedpoints]

                    topoffquery = topoffquery.replace("{dt}","'" + vd + "'")
                else:
                    topoffquery = topoffquery.replace("{dt}",vd)

            self.core.logger.debug("Upper Edge Query #2: " + topoffquery)
            #Now get all of the records that match
            try:
                csr = db.cursor()
                csr.execute(topoffquery);
                for row in csr:
                    query.AddLine(str(row[1]),row[0],row[2])
            except:
                traceback.print_exc()

        elapsed = (datetime.datetime.now() - stime).total_seconds()

        self.core.logger.debug("Total Time After Lower Edge - " + str(elapsed))

        db.close()
        
        return query.Finish()

class driverfactory:
    def createinstance(self):
        return ardihistdriver()
            
if __name__ == "__main__":
    sdf = driverfactory()
    base = histdriverbase.historian()
    base.start(sdf)